/*
 * milter-ahead.c
 *
 * Copyright 2004 by Anthony Howe. All rights reserved.
 *
 * The following should be added to the sendmail.mc file:
 *
 *	INPUT_MAIL_FILTER(
 *		`milter-ahead',
 *		`S=unix:/var/lib/milter-ahead.socket, T=C:1m;S:30s;R:6m;E:1m'
 *	)dnl
 */

/***********************************************************************
 *** Leave this header alone. Its generated from the configure script.
 ***********************************************************************/

#include "config.h"

/***********************************************************************
 *** You can change the stuff below if the configure script doesn't work.
 ***********************************************************************/

#ifndef RUN_AS_USER
#define RUN_AS_USER			"smmsp"
#endif

#ifndef RUN_AS_GROUP
#define RUN_AS_GROUP			"smmsp"
#endif

#ifndef STATEDIR
#define STATEDIR			"/var/lib/" MILTER_NAME
#endif

#ifndef SMTP_PORT
#define SMTP_PORT			25
#endif

#ifndef CACHE_FILE
#define CACHE_FILE			STATEDIR "/cache.db"
#endif

/*
 * Prefered cache implementation when available is Berkeley DB.
 *
 *	NULL		(default)
 * 	"bdb"		(memory/disk, persistant)
 *	"hash"		(in memory, not persistant across restarts)
 *	"flatfile"	(not recommend for performance)
 */
#ifndef CACHE_TYPE
#define CACHE_TYPE			NULL
#endif

/*
 * Expire entries in the cache every N connections.
 */
#ifndef DEFAUL_GC_FREQUENCY
#define DEFAUL_GC_FREQUENCY		1000
#endif

#ifndef DEFAULT_SOCKET_TIMEOUT
#define DEFAULT_SOCKET_TIMEOUT		120
#endif

#ifndef DEFAULT_CACHE_TTL
#define DEFAULT_CACHE_TTL		(7 * 24 * 3600)
#endif

#define NON_BLOCKING_READ
#define NON_BLOCKING_WRITE

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

/* Re-assert this macro just in case. May cause a compiler warning. */
#define _REENTRANT	1

#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/stat.h>

#include <com/snert/lib/version.h>

#if LIBSNERT_MAJOR < 1 || LIBSNERT_MINOR < 39
# error "LibSnert/1.39 or better is required"
#endif

#if ! defined(HAVE_GMTIME_R) || ! defined(HAVE_LOCALTIME_R)
# error "thread safe gmtime_r() and localtime_r() are required"
#endif

#if defined(LIBSNERT_WITH_BERKELEY_DB) && defined(MILTER_WITH_BERKELEY_DB)
# define HAVE_BERKELEY_DB
#else
# undef HAVE_BERKELEY_DB
#endif

#include <com/snert/lib/mail/smf.h>
#include <com/snert/lib/io/Socket.h>
#include <com/snert/lib/sys/Mutex.h>
#include <com/snert/lib/util/Cache.h>
#include <com/snert/lib/util/Text.h>
#include <com/snert/lib/util/getopt.h>

/***********************************************************************
 *** Constants
 ***********************************************************************/

#define PARSE_MAILER_HOST	"[%255[^]]]"		/* DOMAIN_LENGTH = 255 */

#define HOP_INFO_FORMAT		"server %s for <%s>"
#define HOP_INFO_ARGS		rcpt_host, rcpt_addr

#define	TAG_FORMAT		"%05d %s: "
#define	TAG_ARGS		data->work.cid, data->work.qid

/***********************************************************************
 *** Global Variables
 ***********************************************************************/

typedef struct {
	sfsistat status;
	long stamp;
	long count;
} CacheEntry;

static CacheEntry cacheUndefinedEntry = { X_SMFIS_UNKNOWN, 0, 0 };
static long cacheTTL = DEFAULT_CACHE_TTL;
static volatile Cache cache;
static char *cacheType = CACHE_TYPE;
static char *cacheFile = CACHE_FILE;

typedef struct {
	smfWork work;
	Socket server;				/* per RCPT */
	char line[SMTP_LINE_SIZE+1];		/* general purpose */
} *workspace;

static smfInfo milter = {
	MILTER_NAME,
	MILTER_VERSION,
	MILTER_COPYRIGHT,
	RUN_AS_USER,	
	RUN_AS_GROUP,
	STATEDIR,
	NULL,
	SMF_STDIO_CLOSE
};

static volatile long connectionCount;

static long gcFrequency = DEFAUL_GC_FREQUENCY;
static long socketTimeout = DEFAULT_SOCKET_TIMEOUT;
static char *forceNextHop;
static char *sendmailCf;
static int isBackupMx;

static char getoptString[] = "bc:f:g:hn:t:T:v:z:";

static char usageMessage[] =
"\n"
"milter-ahead [-b][-c ttl][-f cf][-g freq][-h][-n host][-t timeout]\n"
"             [-T timeout][-v detail][-z type[,file]] socket\n"
"\n"
"-b\t\tfor a backup-MX, accept mail when the primary is not reachable\n"
"-c ttl\t\tcache responses for ttl seconds\n"
"-f cf\t\tpathname of sendmail.cf file\n"
"-g freq\t\tcache garbadge collection frequency\n"
"-h\t\tthis usage summary\n"
"-n host\t\tforce call-ahead to this host, overriding ${rcpt_host}\n"
"-t timeout\tsendmail/milter I/O timeout in seconds, 0 indefinite\n"
"-T timeout\tmilter/server I/O timeout in seconds, 0 indefinite\n"
"-v detail\tcomma separated list of what to write to the mail log:\n"
"\t\t0, all, info, trace, parse, debug, dialog, state, dns\n"
"\t\tcache, database, socket-fd, socket-all, libmilter\n"
"-z type[,file]\tcache type: bdb (default), flatfile, hash. An optional\n"
"\t\tfile path can be specified for bdb or flatfile types.\n"
"\n"
MILTER_VERSION " " MILTER_COPYRIGHT "\n"
"\n"
;

/***********************************************************************
 *** Support routines.
 ***********************************************************************/

/*
 * Print a line to the remote SMTP server. Return true
 * on success, otherwise false for a write error.
 */
int
printline(workspace data, char *line)
{
	int error;
	int rc = 0;

	if (smfLogDetail & SMF_LOG_DIALOG)
		syslog(LOG_DEBUG, TAG_FORMAT "> %s", TAG_ARGS, line);

#if defined(NON_BLOCKING_WRITE)
# if ! defined(NON_BLOCKING_READ)
	/* Do not block on sending to the SMTP server just yet. */
	(void) SocketSetNonBlocking(data->server, 1);
# endif

	if (SocketPrint(data->server, line) < 0) {
		error = SocketGetError(data->server);
		syslog(LOG_ERR, TAG_FORMAT "SocketPrint() error: %s (%d)", TAG_ARGS, strerror(error), error);
		goto error0;
	}

	/* Now wait for the output to be sent to the SMTP server. */
	if (SocketWaitForOutput(data->server, socketTimeout) != 0) {
		syslog(LOG_ERR, TAG_FORMAT "timeout before output sent to SMTP server", TAG_ARGS);
		goto error0;
	}
#else
	if (SocketPrint(data->server, line) < 0) {
		error = SocketGetError(data->server);
		syslog(LOG_ERR, TAG_FORMAT "SocketPrint() error: %s (%d)", TAG_ARGS, strerror(error), error);
		goto error0;
	}
#endif
	rc = 1;
error0:
#if ! defined(NON_BLOCKING_READ) && defined(NON_BLOCKING_WRITE)
	(void) SocketSetNonBlocking(data->server, 0);
#endif
	return rc;
}

int
printlines(workspace data, char **lines) 
{
        for ( ; *lines != (char *) 0; lines++) {
        	if (!printline(data, *lines))
        		return 0;
        }
        
        return 1;
}      

int
getSmtpResponse(workspace data, int code, char *line, int linesize, int *status)
{
	int error;
	char *stop;
	long length, value;

	if (linesize < sizeof (long)) {
		syslog(LOG_CRIT, TAG_FORMAT "SMTP line buffer too small (%d)", TAG_ARGS, linesize);
		exit(1);
	}
	
#if defined(NON_BLOCKING_READ) && ! defined(NON_BLOCKING_WRITE)
	/* Do not block on reading from the SMTP server just yet. */
	(void) SocketSetNonBlocking(data->server, 1);
#endif

	/* Ideally we should collect _all_ the response lines into a variable
	 * length buffer (see com/snert/lib/util/Buf.h), but I don't see any
	 * real need for it just yet.
	 */

	do {
		/* Reset this return code for each input line of a
		 * multiline response in case there is a read error.
		 */
		value = 450;
		stop = line;

		/* Erase the first 4 bytes of the line buffer, which
		 * corresponds with the 3 ASCII digit status code
		 * followed by either a ASCII hyphen or space.
		 */
		line[0] = line[1] = line[2] = line[4] = '\0';

		error = SocketWaitForInput(data->server, socketTimeout);

		if (error == ETIMEDOUT) {
			syslog(LOG_ERR, TAG_FORMAT "timeout before input from SMTP server", TAG_ARGS);
			goto error0;
		}

		if (error != 0) {
			syslog(LOG_ERR, TAG_FORMAT "input error from SMTP server: %s (%d)", TAG_ARGS, strerror(error), error);
			goto error0;
		}

		length = SocketReadLine(data->server, line, linesize);

		if ((error = SocketGetError(data->server)) != 0 || length == -1) {
			syslog(LOG_ERR, TAG_FORMAT "read error from SMTP server: %s (%d)", TAG_ARGS, strerror(error), error);
			goto error0;
		}

		if (length == -2) {
			syslog(LOG_ERR, TAG_FORMAT "unexpected EOF from SMTP server", TAG_ARGS);
			goto error0;
		}

		/* Did we read sufficient characters for a response code? */
		if (length < 4) {
			syslog(LOG_ERR, TAG_FORMAT "truncated response from SMTP server, length=%ld", TAG_ARGS, length);
			goto error0;
		}

		if (smfLogDetail & SMF_LOG_DIALOG)
			syslog(LOG_DEBUG, TAG_FORMAT "< %s", TAG_ARGS, line);

		if ((value = strtol(line, &stop, 10)) == 421) {
			if (smfLogDetail & SMF_LOG_DIALOG)
				syslog(LOG_DEBUG, TAG_FORMAT "SMTP server closing connection", TAG_ARGS);
			break;
		}
	} while (line[3] == '-');
error0:
#if defined(NON_BLOCKING_READ) && ! defined(NON_BLOCKING_WRITE)
	(void) SocketSetNonBlocking(data->server, 0);
#endif
	if (status != (int *) 0)
		*status = (int) value;

	return value == code && line + 3 == stop;
}

/***********************************************************************
 *** Cache Functions
 ***********************************************************************/

int
cacheGet(workspace data, char *name, CacheEntry *entry)
{
	Data value;
	struct data key;

	*entry = cacheUndefinedEntry;	
	DataInitWithBytes(&key, name, strlen(name)+1);

	if (smfLogDetail & SMF_LOG_CACHE)
		syslog(LOG_DEBUG, "cache get {%s}", name);

	if (pthread_mutex_lock(&smfMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex lock in cacheGet() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);	

	value = cache->get(cache, &key);

	if (pthread_mutex_unlock(&smfMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex unlock in cacheGet() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);	

	if (value == NULL)
		return -1;
	
	*entry = *(CacheEntry *)(value->base(value));
	value->destroy(value);

	if (smfLogDetail & SMF_LOG_CACHE)
		syslog(LOG_DEBUG, TAG_FORMAT "cacheGet(%lx, '%s', {st=%d, cn=%ld}) ", TAG_ARGS, (long) data, name, entry->status, entry->count);

	return 0;
}

int
cachePut(workspace data, char *name, CacheEntry *entry)
{
	int rc;
	struct data key, value;

	DataInitWithBytes(&key, name, strlen(name)+1);
	DataInitWithBytes(&value, (unsigned char *) entry, sizeof (*entry));

	if (smfLogDetail & SMF_LOG_CACHE)
		syslog(LOG_DEBUG, "cache put {%s}", name);
	
	if (pthread_mutex_lock(&smfMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex lock in cachePut() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);

	rc = cache->put(cache, &key, &value);
	
	if (pthread_mutex_unlock(&smfMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex unlock in cachePut() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);	

	return rc;
}

int
cacheExpireEntries(void *key, void *value, void *data)
{
	time_t *now = data;
	CacheEntry *entry = (CacheEntry *) ((Data) value)->base(value);
	
	if (*now < entry->stamp + cacheTTL)
		return 1;

	if (smfLogDetail & SMF_LOG_CACHE)
		syslog(LOG_DEBUG, "deleting <%s>", ((Data) key)->base(key));

	return -1;	
}

int
cacheGarbageCollect(workspace data)
{
	time_t now = time(NULL);
	
	if (pthread_mutex_lock(&smfMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex lock in cacheGarbageCollect() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);		

	connectionCount++;
	
	if (smfLogDetail & SMF_LOG_CACHE)
		syslog(LOG_DEBUG, "%ld connections", connectionCount);

	if (gcFrequency <= connectionCount) {
		if (smfLogDetail & SMF_LOG_CACHE)
			syslog(LOG_DEBUG, "garbage collecting cache");

		cache->walk(cache, cacheExpireEntries, &now);
		connectionCount = 0;

		if (smfLogDetail & SMF_LOG_CACHE)
			syslog(LOG_DEBUG, "syncing cache");

		if (cache->sync(cache))
			syslog(LOG_ERR, "cache sync error: %s (%d)", strerror(errno), errno);
	}

	if (pthread_mutex_unlock(&smfMutex))
		syslog(LOG_ERR, TAG_FORMAT "mutex unlock in cacheGarbageCollect() failed: %s (%d) ", TAG_ARGS, strerror(errno), errno);	

	return 0;
}

/***********************************************************************
 *** Handlers
 ***********************************************************************/

/*
 * Open and allocate per-connection resources.
 */
static sfsistat
filterOpen(SMFICTX *ctx, char *client_name, _SOCK_ADDR *raw_client_addr)
{
	workspace data;
	unsigned short cid;
	char client_addr[5+IPV6_LENGTH+1];	
	
	cid = smfOpenProlog(ctx, client_name, raw_client_addr, client_addr, sizeof (client_addr));	
	
	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "filterOpen(%lx, '%s', [%s])", cid, smfNoQueue, (long) ctx, client_name, client_addr);

	if ((data = calloc(1, sizeof *data)) == NULL)
		goto error0;

	if (smfi_setpriv(ctx, (void *) data) == MI_FAILURE) {
		syslog(LOG_ERR, TAG_FORMAT "failed to save workspace", TAG_ARGS);
		goto error1;
	}

	data->work.ctx = ctx;
	data->work.cid = cid;
	data->work.qid = smfNoQueue;

	return SMFIS_CONTINUE;
error1:
	free(data);
error0:
	return smfNullWorkspaceError("filterOpen");
}

static sfsistat
filterRcpt(SMFICTX *ctx, char **args)
{
	int status;
	sfsistat rc;
	workspace data;
	CacheEntry entry;
	char *error, *auth_authen, *rcpt_addr, *rcpt_host, *rcpt_mailer, *helo;

	rc = SMFIS_CONTINUE;

	if ((data = (workspace) smfi_getpriv(ctx)) == NULL)
		return smfNullWorkspaceError("filterRcpt");

	free(data->work.rcpt);
	data->work.rcpt = NULL;
	
	if ((data->work.qid = smfi_getsymval(ctx, "i")) == NULL)
		data->work.qid = smfNoQueue;
		
	rcpt_addr = smfi_getsymval(ctx, "{rcpt_addr}");
	rcpt_host = smfi_getsymval(ctx, "{rcpt_host}");
	rcpt_mailer = smfi_getsymval(ctx, "{rcpt_mailer}");
	auth_authen = smfi_getsymval(ctx, "{auth_authen}");

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(
			LOG_DEBUG, TAG_FORMAT "RCPT='%s' rcpt_addr='%s' rcpt_host='%s' rcpt_mailer='%s'",
			TAG_ARGS, args[0], rcpt_addr, rcpt_host, rcpt_mailer
		);

	/* BY-PASS if the mailer is local. */
	if (rcpt_host == NULL || *rcpt_host == '\0') {
		if (smfLogDetail & (SMF_LOG_INFO|SMF_LOG_TRACE))
			syslog(LOG_INFO, TAG_FORMAT "looks like a local recipient, skipping", TAG_ARGS);
		return SMFIS_CONTINUE;
	}

	/* BY-PASS if the MX needs to be looked up. */
	if (*rcpt_host != '[') {
		if (smfLogDetail & (SMF_LOG_INFO|SMF_LOG_TRACE))
			syslog(LOG_INFO, TAG_FORMAT "rcpt_host='%s' is not a defined route, skipping", TAG_ARGS, rcpt_host);
		return SMFIS_CONTINUE;
	}
	
	/* BY-PASS if the sender is authenticated. */
	if (auth_authen != NULL) {
		if (smfLogDetail & (SMF_LOG_INFO|SMF_LOG_TRACE))
			syslog(LOG_INFO, TAG_FORMAT "authenticated sender \"%s\", skipping", TAG_ARGS, auth_authen);
		return SMFIS_CONTINUE;
	}
	
	/* TEST: Validate the syntax and chop up the RCPT address. */
	if ((error = (char *) parsePath(args[0], 0, &data->work.rcpt)) != NULL)
		return smfReply(&data->work, 553, (char *) 0, error);

	if (0 < cacheTTL && cacheGet(data, rcpt_addr, &entry) == 0) {
		/* BY-PASS if the recipient is address is known. The cache entry
		 * is not touched and allowed to expire.
		 *
		 * NOTE is might be better to remove the REJECT and TEMPFAIL
		 * case and perform the call-ahead just in case the address is
		 * recreated.
		 */
		switch (entry.status) {
		case SMFIS_ACCEPT:		
		case SMFIS_CONTINUE:		
			if (time(NULL) < entry.stamp + cacheTTL) {
				if (smfLogDetail & (SMF_LOG_INFO|SMF_LOG_TRACE))
					syslog(LOG_INFO, TAG_FORMAT "recipient <%s> (%d) cached, skipping", TAG_ARGS, rcpt_addr, entry.status);
				return SMFIS_CONTINUE;
			}
			break;
		case SMFIS_REJECT:
			if (time(NULL) < entry.stamp + cacheTTL) {
				return smfReply(&data->work, 550, NULL, "User unknown");
			}
			break;

		case SMFIS_TEMPFAIL:
			if (time(NULL) < entry.stamp + cacheTTL) {
				return smfReply(&data->work, 450, NULL, "Recipient denied");
			}
			break;
		}
	}

	if (smfLogDetail & SMF_LOG_PARSE)
		syslog(
			LOG_DEBUG, TAG_FORMAT "address='%s' localleft='%s' localright='%s' domain='%s'",
			TAG_ARGS, data->work.rcpt->address.string, data->work.rcpt->localLeft.string,
			data->work.rcpt->localRight.string, data->work.rcpt->domain.string
		);

	if (*data->work.rcpt->localLeft.string == '\0')
		return smfReply(&data->work, 553, "5.1.0", "cannot deliver to null address");

	if (smfLogDetail & SMF_LOG_DIALOG)
		syslog(LOG_DEBUG, TAG_FORMAT "contacting " HOP_INFO_FORMAT, TAG_ARGS, HOP_INFO_ARGS);

	/* Connect to next hop. */
	if (forceNextHop == NULL) {
		(void) sscanf(rcpt_host, PARSE_MAILER_HOST, data->line);
	} else {
		(void) snprintf(data->line, sizeof (data->line), "%s", forceNextHop);
		rcpt_host = forceNextHop;
	}
	
	data->server = SocketOpenTcpClientWait(data->line, SMTP_PORT, socketTimeout);

	if (data->server == NULL) {
		if (isBackupMx)
			rc = SMFIS_CONTINUE;
		else
			rc = smfReply(&data->work, 450, "4.7.1", "no connection with " HOP_INFO_FORMAT, HOP_INFO_ARGS);
		goto error0;
	}
	
#if defined(NON_BLOCKING_READ) && defined(NON_BLOCKING_WRITE)
	/* Set non-blocking I/O once for both getSmtpResponse() and
	 * printline() and leave it that way. 
	 */
	if (SocketSetNonBlocking(data->server, 1)) {
		syslog(LOG_ERR, TAG_FORMAT "internal error: SocketSetNonBlocking() failed: %s (%d)", TAG_ARGS, strerror(errno), errno);
		rc = smfReply(&data->work, 451, "4.4.3", "failed to set non-blocking socket");
		goto error1;
	}
#endif
	/* Read welcome response. */
	if (!getSmtpResponse(data, 220, data->line, SMTP_LINE_SIZE, &status)) {
		/* TEST if we got "554 No SMTP service here" response for the next hop? */
		if (SMTP_IS_PERM(status)) {
			rc = smfReply(&data->work, 550, NULL, "not welcomed by " HOP_INFO_FORMAT ". Their reason \"%s\"", HOP_INFO_ARGS, data->line);
			goto error1;
		}
	
		rc = smfReply(&data->work, 450, "4.7.1", "no welcome from " HOP_INFO_FORMAT, HOP_INFO_ARGS);
		goto error1;
	}

	/* TEST: Send our hello. Failure to accept us indicates an
	 * unwillingness to accept any mail from us. Return the
	 * favour.
	 */
	status = 450;

	if ((helo = smfi_getsymval(ctx, "{if_name}")) == NULL
	&&  (helo = smfi_getsymval(ctx, "{if_addr}")) == NULL
	&&  (helo = smfi_getsymval(ctx, "j")) == NULL)
		helo = "127.0.0.1";
		
	(void) snprintf(data->line, sizeof (data->line), "HELO %s\r\n", helo);
	if (!printline(data, data->line)
	|| !getSmtpResponse(data, 250, data->line, SMTP_LINE_SIZE, &status)) {
		rc = smfReply(&data->work, status, NULL, HOP_INFO_FORMAT " did not accept HELO", HOP_INFO_ARGS);
		goto error1;
	}

	/* TEST: Use the DSN address to verify that they're playing
	 * by the rules. Also using the DSN address avoids loops
	 * in the case of both systems using milter-sender.
	 */
	status = 450;
	if (!printline(data, "MAIL FROM:<>\r\n")
	|| !getSmtpResponse(data, 250, data->line, SMTP_LINE_SIZE, &status)) {
		rc = smfReply(&data->work, status, NULL, HOP_INFO_FORMAT " does not accept <> address as required by RFC 821, 1123, 2505, and 2821", HOP_INFO_ARGS);
		goto error1;
	}

	/* TEST: Finally see if they're willing to accept email for
	 * the recipient. An error here probably means an user unknown,
	 * mailbox full, or similar nonsense, in which case we don't
	 * want to deal with them.
	 */
	(void) snprintf(data->line, SMTP_LINE_SIZE, "RCPT TO:<%s>\r\n", rcpt_addr);

	status = 450;
	if (!printline(data, data->line)
	|| !getSmtpResponse(data, 250, data->line, SMTP_LINE_SIZE, &status)) {
		/* Skip leading status and extended status codes */
		int len = strspn(data->line, "0123456789. ");
		rc = smfReply(&data->work, status, NULL, HOP_INFO_FORMAT " rejected recipient. Their reason \"%s\"", HOP_INFO_ARGS, data->line + len);
		goto error1;
	}

	if (smfLogDetail & SMF_LOG_DIALOG)
		syslog(LOG_DEBUG, TAG_FORMAT "" HOP_INFO_FORMAT " accepted recipient", TAG_ARGS, HOP_INFO_ARGS);
		
	rc = SMFIS_CONTINUE;
error1:
	/* Cache the result of the call ahead (success or failure). */
	entry.status = rc;
	entry.stamp = time(NULL);
	entry.count = 1;

	if (cachePut(data, rcpt_addr, &entry)) {
		/* We can survive without the cache, though performance
		 * may degrade, since we will retest each message from
		 * this sender instead.
		 */
		syslog(LOG_WARNING, TAG_FORMAT "failed to add <%s> to cache", TAG_ARGS, rcpt_addr);
	}	

	if (data->server != (Socket) 0) {			
		/* Some how we are getting ENOTSOCK during getSmtpResponse(),
		 * probably from a dropped connection or some such. However
		 * attempting then to send the QUIT appears to crash the
		 * milter. So now if there is any sort of error, we drop
		 * the connection hard, instead of politely.
		 */
		if (SocketGetError(data->server) == 0) {
			(void) printline(data, "QUIT\r\n");

			/* Don't wait to read a response from the QUIT,
			 * since some SMTP servers drop the connection
			 * without sending the response.
			 */

			(void) SocketSetLinger(data->server, 3);

			if (smfLogDetail & SMF_LOG_DIALOG)
				syslog(LOG_DEBUG, TAG_FORMAT "closing SMTP connection", TAG_ARGS);
		}

		SocketClose(data->server);
		data->server = NULL;
	}
error0:	
	return rc;
}

/*
 * Close and release per-connection resources.
 */
static sfsistat
filterClose(SMFICTX *ctx)
{
	workspace data;
	unsigned short cid;

	if ((data = (workspace) smfi_getpriv(ctx)) == NULL)
		return smfNullWorkspaceError("filterClose");

	cid = smfCloseEpilog(&data->work);
	free(data);

	if (smfLogDetail & SMF_LOG_TRACE)
		syslog(LOG_DEBUG, TAG_FORMAT "filterClose(%lx)", cid, smfNoQueue, (long) ctx);

	return SMFIS_CONTINUE;
}

/***********************************************************************
 ***  Milter Definition Block
 ***********************************************************************/

static struct smfiDesc smfilter = {
	MILTER_NAME,		/* filter name */
	SMFI_VERSION,		/* version code -- do not change */
	0,			/* flags */
	filterOpen,		/* connection info filter */
	NULL,			/* SMTP HELO command filter */
	NULL,			/* envelope sender filter */
	filterRcpt,		/* envelope recipient filter */
	NULL,			/* header filter */
	NULL,			/* end of header */
	NULL,			/* body block filter */
	NULL,			/* end of message */
	NULL,			/* message aborted */
	filterClose,		/* connection cleanup */
};

/***********************************************************************
 *** Startup
 ***********************************************************************/

static void
atExitCleanUp()
{
	cache->destroy(cache);
	smfAtExitCleanUp();
}

int
main(int argc, char **argv)
{
	int i, ch;

	openlog(MILTER_NAME, LOG_PID, LOG_MAIL);
	(void) smfi_settimeout(SMF_SOCKET_TIMEOUT);

	smfSetProcessOwner(&milter);

	while ((ch = getopt(argc, argv, getoptString)) != -1) {
		switch (ch) {
		case 'b':
			isBackupMx = 1;
			break;
		case 'c':
			cacheTTL = strtol(optarg, (char **) 0, 10);
			break;
		case 'f':
			sendmailCf = optarg;
			break;
		case 'g':
			gcFrequency = strtol(optarg, (char **) 0, 10);
			break;			
		case 'n':
			forceNextHop = optarg;
			break;
		case 'T':
			socketTimeout = strtol(optarg, (char **) 0, 10);
			break;
		case 't':
			(void) smfi_settimeout((int) strtol(optarg, NULL, 10));
			break;
		case 'v':
			(void) smfSetLogDetail(optarg);
			break;
		case 'z':
			cacheType = optarg;
			i = strcspn(cacheType, ",");
			if (cacheType[i] == ',') {
				cacheFile = &cacheType[i + 1];
				cacheType[i] = '\0';
			}
			break;
		default:
			syslog(LOG_ERR, "unknown command line option -%c", optopt);
		case 'h':
		case '?':
			(void) fprintf(stderr, usageMessage);
			return 2;
		}
	}

	if (optind + 1 != argc) {
		syslog(LOG_ERR, "missing command line argument");
		(void) fprintf(stderr, usageMessage);
		return 2;
	}

	milter.socket = argv[optind];	

	/* Socket timeout is specified in seconds on the command line, 
	 * but we need milliseconds for the function calls.
	 */
	socketTimeout *= 1000;
	if (socketTimeout < 0)
		socketTimeout = 0;

	if ((cache = CacheCreate(cacheType, cacheFile)) == NULL) {
		syslog(LOG_ERR, "failed to create cache\n");
		return 1;
	}
	
	if (atexit(atExitCleanUp)) {
		syslog(LOG_ERR, "atexit() failed\n");
		return 1;
	}	

	if (smfLogDetail & (SMF_LOG_SOCKET_FD|SMF_LOG_SOCKET_ALL))
		SocketSetDebugMask(
			(smfLogDetail & SMF_LOG_SOCKET_ALL) 
				? SOCKET_DEBUG_ALL
				: SOCKET_DEBUG_FD
		);

	if (SocketInit()) {
		syslog(LOG_ERR, "SocketInit() error\n");
		return 1;
	}
	
	return smfMainStart(&milter, smfilter);
}
